require 'json'
require 'uu/os/lang/json_patch'
require 'uu/os/http/http_method'

module UU
  module OS
    module HTTP

      # POST Method.
      class HTTPPost < UU::OS::HTTP::HTTPMethod

        # Creates new instance of HTTPGet
        # @param [UU::OS::HTTP::HTTPClient] client Instance of HTTPClient
        # @param [String] path Request path
        # @private
        def initialize(client, path) # :nodoc:
          super(client, path, :post)
        end

        # Sets custom Accept header.
        # @param [Array<String>] content_type Accepted content-type.
        # @return [HTTPMethod] Self instance for configuration chaining.
        def accept(*content_type)
          headers[:accept] = content_type
          return self
        end

        # Defines content-type to be used for payload.
        # @param [String] content_type Content-type
        # @return [HTTPPost] Self instance for configuration chaining.
        def content_type(content_type)
          headers[:content_type] = content_type
          return self
        end

        # Executes method sending given payload. In case you need to stream response or upload
        # large data, {#execute_async execute_async} should be used.
        # @param [Hash, Array, IO, String, Object] payload Payload to be sent. If no {#content_type content-type} was set, Hash is automatically sent as +multipart/form-data+.
        #    If {#content_type content-type} was set as +application/json+, any object with exclusion of String and IO is transformed using {Object#to_json #to_json} method.
        # @return [UU::OS::HTTP::HTTPResponse] Response object.
        def execute(payload = nil)
          super(process_payload(payload))
        end

        # Executes method sending given payload asynchronously. Unlike {#execute execute}, this call is non blocking.
        # @param [Hash, Array, IO, String, Object] payload Payload to be sent. If no {#content_type content-type} was set, Hash is automatically sent as +multipart/form-data+.
        #    If {#content_type content-type} was set as +application/json+, any object with exclusion of String and IO is transformed using {Object#to_json #to_json} method.
        # @return [UU::OS::HTTP::HTTPResponse] Response object.
        def execute_async(payload = nil)
          super(process_payload(payload))
        end

        private

        # Tries to process any type of payload so that it can be transferred via HTTP protocol.
        # @param [Object] payload Payload to be processed
        # @return [Object] Processed payload
        # @private
        def process_payload(payload) # :nodoc:
          return payload if payload.nil?
          if !headers.has_key?(:content_type)
            if payload.kind_of?(Hash)
              headers[:content_type] = 'multipart/form-data'
              return prepare_multipart(payload)
            end
          else
            if headers.contains(:content_type, 'application/json')
              if !payload.kind_of?(String) && !payload.kind_of?(IO) && !payload.kind_of?(StringIO) && !payload.kind_of?(Tempfile)
                return payload.to_json
              end
            elsif headers.contains(:content_type, 'multipart/form-data') && payload.kind_of?(Hash)
              return prepare_multipart(payload)
            end
          end
          if !payload.kind_of?(String) && !payload.kind_of?(IO) && !payload.kind_of?(StringIO) && !payload.kind_of?(Tempfile)
            raise TypeError, "not obvious how to send #{payload.class} as #{headers.has_key?(:content_type) ? headers[:content_type] : 'application/octet-stream'}"
          end
          return payload
        end

        # Transforms Hash payload into multipart payload.
        # @param [Hash] payload Payload in key-value hash format
        # @return [Array<Hash>] Multipart payload
        # @private
        def prepare_multipart(payload) # :nodoc:
          multi_part = []
          payload.each do |key, value|
            part = {:'content-disposition' => "form-data; name=\"#{key}\""}
            if value.kind_of?(Hash) || value.kind_of?(Array)
              part[:'content-type'] = 'application/json'
              part[:content] = value.to_json
            elsif value.kind_of?(Numeric) || value.kind_of?(TrueClass) || value.kind_of?(FalseClass)
              # Single value can be sent as JSON according to new RFC (http://rfc7159.net).
              # This way we can correctly deserialize boolean and numbers on server while
              # it remains as plain string for servers which ignore custom part header.
              part[:'content-type'] = 'application/json'
              part[:content] = value.to_json
            elsif value.kind_of?(Date) || value.kind_of?(Time)
              # JSON does not have own format for Date/Time so we can serialize them as
              # plain string (even if server knows this is JSON it will not be able to
              # determine whether it is simple string value or serialized Date/Time).
              part[:'content-type'] = 'application/octet-stream'
              part[:content] = value.to.json[1..-2]
            else
              if value.respond_to?(:content_type)
                part[:'content-type'] = value.content_type
              else
                part[:'content-type'] = 'application/octet-stream'
              end
              if value.respond_to?(:encoding)
                part[:'content-type'] << "; charset=\"#{value.encoding}\""
              end
              # Adding attachment to content-disposition of multipart is not in specification
              # and therefore should not be used (server must not depend on it).
              # if value.respond_to?(:read)
              #   part[:'content-disposition'] << '; attachment'
              # end
              if value.respond_to?(:path)
                part[:'content-disposition'] << "; filename=\"#{File.basename(value.path)}\""
              end
              part[:content] = value
            end
            multi_part << part
          end
          return multi_part
        end

      end

    end
  end
end
