require 'stringio'

module UU
  module OS
    module HTTP

      # Module providing methods for multipart request parsing.
      module MultipartParser

        # Parses multipart body.
        # @param [String, IO] body Multipart request body
        # @param [String] boundary Multipart item boundary
        # @param [Symbol, String] binary_name Name of attribute containing large
        #    binary data. In case attribute with this name is found, its content
        #    is returned as stream and rest of multipart body is dumped (binary
        #    attribute should be always the last item of multipart body).
        # @return [Array<Hash<Symbol,Object>>] Parsed multipart request in for
        #   of +Array+ of +Hashes+. Each +Hash+ represents one form item with
        #   attributes +:name+ (item name), +:filename+ (optional), +:content_type+
        #   (optional) and +:body+ (+String+ or +IO+ based on +binary_name+ parameter).
        def parse_multipart(body, boundary, binary_name = nil)
          MultipartParser.parse_multipart(body, boundary, binary_name)
        end

        # @see #parse_multipart
        # @private
        def self.parse_multipart(body, boundary, binary_name = nil) # :nodoc:
          items = []
          binary_name = binary_name.to_s if !binary_name.nil?
          body = StringIO.new(body.to_s) if !body.respond_to?(:read)
          boundary = dequote(boundary)
          boundary_regex = /^--#{boundary}(--|\r\n)/
          # Multipart item starts with boundary
          line = read_line(body)
          while !line.nil? && line !~ boundary_regex do
            line = read_line(body)
          end
          # Read multipart item
          line = read_line(body)
          while !line.nil? do
            items << (item = {})
            # Read headers
            while !line.nil? && line !~ /^\r\n$/ do
              header = line.strip
              if header =~ /^content-type:(.*)/i
                item[:content_type] = $1.strip
              elsif header =~ /^content-disposition:(.*)/i
                value = $1.strip
                if value =~ /(^|;)[ ]*name=([^;]+)/i
                  item[:name] = dequote($2)
                end
                if value =~ /(^|;)[ ]*filename=([^;]+)/i
                  item[:filename] = dequote($2)
                end
              end
              line = read_line(body)
            end
            # Read body. In case binary attribute is requested, it is
            # returned as stream and any following attributes are dumped
            # (because we cannot store binary data in memory or tempfiles).
            if !item[:name].nil? && binary_name == item[:name]
              item[:body] = read_binary(body, boundary)
              return items
            else
              item[:body] = read_string(body, boundary)
            end
            # Skip empty lines after item boundary to reach next header (or end of stream).
            line = read_line(body)
            while !line.nil? && line =~ /^\r\n$/
              line = read_line(body)
            end
          end
          body.close
          return items
        end

        private

        # @private
        def self.read_line(io) # :nodoc:
          return io.readline("\r\n")
        rescue EOFError
          return nil
        end
        private_class_method :read_line

        # @private
        def self.read_string(io, boundary) # :nodoc:
          boundary_regex = /^--#{boundary}(--|\r\n)/
          value = ''
          prev_line = ''
          line = read_line(io)
          while !line.nil? && !(prev_line =~ /\r\n$/ && line =~ boundary_regex) do
            value << prev_line
            prev_line, line = line, read_line(io)
            tmp = 0
          end
          value << prev_line[0..-3]
          return value
        end
        private_class_method :read_string

        # @private
        def self.read_binary(io, boundary) # :nodoc:
          boundary_regex = /\r\n--#{boundary}(--|\r\n)/
          read, write = IO.pipe
          write.write('')
          Thread.new {
            prev_chunk = ''
            chunk = io.read(32768)
            # Read data until end of stream or boundary is reached
            while !chunk.nil? && (block = ('' << prev_chunk) << chunk) !~ boundary_regex do
              write.write(prev_chunk)
              prev_chunk, chunk = chunk, io.read(32768)
            end
            # Write last block (without boundary)
            write.write(block.split(boundary_regex)[0])
            # Dump rest of stream
            while !io.read(32768).nil? do
              Thread.pass
            end
            io.close()
            write.close()
          }
          return read
        end
        private_class_method :read_binary

        # @private
        def self.dequote(value) # :nodoc:
          value = value.strip
          value = value[1..-2] if value =~ /^".*"$/
          return value
        end
        private_class_method :dequote

      end

    end
  end
end
